use std::{cell::RefCell, rc::Rc};

// TODO: Remove this
use wasm_bindgen::prelude::*;
#[wasm_bindgen]
extern "C" {
    #[wasm_bindgen(js_namespace = console)]
    fn log(s: &str);
}

pub trait Content {
    fn get_name(&self) -> String;
    fn get_path(&self) -> String;
}

pub struct File {
    parent_path: String,
    name: String,
    content: String,
}

impl Content for File {
    fn get_name(&self) -> String {
        self.name.clone()
    }

    fn get_path(&self) -> String {
        if self.parent_path == "/" {
            format!("/{}", &self.name)
        } else {
            format!("{}/{}", &self.parent_path, &self.name)
        }
    }
}

#[allow(unused)]
impl File {
    pub fn new(name: String, content: String, parent_path: String) -> File {
        File {
            name,
            content,
            parent_path,
        }
    }

    pub fn empty_new(name: String, parent_path: String) -> File {
        File {
            name,
            content: String::new(),
            parent_path,
        }
    }

    pub fn read(&self) -> String {
        self.content.clone()
    }

    pub fn write(&mut self, content: String) {
        self.content = content;
    }

    pub fn append(&mut self, content: String) {
        self.content.push_str(&content);
    }

    pub fn rename(&mut self, name: String) {
        self.name = name;
    }
}

pub struct Directory {
    name: String,
    files: RefCell<Vec<Rc<RefCell<File>>>>,
    subdirs: RefCell<Vec<Rc<RefCell<Directory>>>>,
    pub parent_path: String,
}

impl Content for Directory {
    fn get_name(&self) -> String {
        self.name.clone()
    }

    fn get_path(&self) -> String {
        if self.parent_path == "/" {
            format!("/{}", &self.name)
        } else {
            format!("{}/{}", &self.parent_path, &self.name)
        }
    }
}

#[allow(unused)]
impl Directory {
    pub fn new(name: String, parent_path: String) -> Directory {
        Directory {
            name,
            files: RefCell::new(Vec::new()),
            subdirs: RefCell::new(Vec::new()),
            parent_path,
        }
    }

    pub fn add_file(&mut self, file: File) -> Result<(), &'static str>{
        if self.get_dir(file.name.clone()).is_some() {
            return Err("File already exists");
        }
        if file.name.contains('/') || file.name.contains('\\')|| file.name.contains(' ') {
            return Err("File name cannot contain spraces, '/' or '\\'");
        }
        self.files.borrow_mut().push(Rc::new(RefCell::new(file)));

        Ok(())
    }

    pub fn add_dir(&mut self, dir: Directory) -> Result<(), &'static str> {
        if self.get_dir(dir.name.clone()).is_some() {
            return Err("Directory already exists");
        }
        if dir.name.contains('/') || dir.name.contains('\\')|| dir.name.contains(' ') {
            return Err("Directory name cannot contain spraces, '/' or '\\'");
        }

        self.subdirs.borrow_mut().push(Rc::new(RefCell::new(dir)));
        Ok(())
    }

    pub fn mkdir(&self, name: String) -> Result<Rc<RefCell<Directory>>, &'static str> {
        if self.get_dir(name.clone()).is_some() {
            return Err("Directory already exists");
        }
        if name.contains('/') || name.contains('\\')|| name.contains(' ') {
            return Err("Directory name cannot contain spraces, '/' or '\\'");
        }

        let dir = Rc::new(RefCell::new(Directory::new(name, self.get_path())));
        self.subdirs.borrow_mut().push(dir.clone());
        Ok(dir)
    }

    pub fn get_file(&self, name: String) -> Option<Rc<RefCell<File>>> {
        for file in self.files.borrow().iter() {
            if file.borrow().get_name() == name {
                return Some(Rc::clone(file));
            }
        }
        None
    }
    
    pub fn get_files(&self) -> Vec<Rc<RefCell<File>>> {
        self.files.borrow().clone()
    }

    pub fn get_subdirs(&self) -> Vec<Rc<RefCell<Directory>>> {
        self.subdirs.borrow().clone()
    }    

    pub fn get_dir(&self, name: String) -> Option<Rc<RefCell<Directory>>> {
        for dir in self.subdirs.borrow().iter() {
            if dir.borrow().get_name() == name {
                return Some(Rc::clone(dir));
            }
        }
        None
    }

    /*pub fn get_file_mut(&self, name: String) -> Option<Rc<File>> {
        for file in self.files.borrow_mut().iter() {
            if file.get_name() == name {
                return Some(Rc::clone(file));
            }
        }
        None
    }

    pub fn get_dir_mut(&self, name: String) -> Option<Rc<Directory>> {
        for dir in self.subdirs.borrow().iter() {
            if dir.get_name() == name {
                return Some(Rc::clone(dir));
            }
        }
        None
    }*/

    pub fn remove_file(&self, name: String) -> Result<(), &'static str> {
        let mut index = 0;
        for file in self.files.borrow().iter() {
            if file.borrow().get_name() == name {
                self.files.borrow_mut().remove(index);
                return Ok(());
            }
            index += 1;
        }
        Err("File not found")
    }

    pub fn remove_dir(&self, name: String) -> Result<(), &'static str> {
        let mut index = 0;
        for dir in self.subdirs.borrow().iter() {
            if dir.borrow().get_name() == name {
                self.subdirs.borrow_mut().remove(index);
                return Ok(());
            }
            index += 1;
        }
        Err("Directory not found")
    }

    pub fn rename(&mut self, name: String) {
        self.name = name;
    }

    pub fn touch(&self, name: String) -> Result<Rc<RefCell<File>>, &'static str> {
        if self.get_file(name.clone()).is_some() {
            return Err("File already exists");
        }
        if name.contains('/') || name.contains('\\')|| name.contains(' ') {
            return Err("File name cannot contain spraces, '/' or '\\'");
        }

        let file = Rc::new(RefCell::new(File::empty_new(name,  self.get_path())));
        self.files.borrow_mut().push(file.clone());
        Ok(file)
    }

    pub fn create_file(&self, name: String, content: String) -> Result<Rc<RefCell<File>>, &'static str> {
        if self.get_file(name.clone()).is_some() {
            return Err("File already exists");
        }
        if name.contains('/') || name.contains('\\')|| name.contains(' ') {
            return Err("File name cannot contain spraces, '/' or '\\'");
        }

        let file = Rc::new(RefCell::new(File::new(name, content, self.get_path())));
        self.files.borrow_mut().push(file.clone());
        Ok(file)
    }
}

impl Drop for Directory {
    fn drop(&mut self) {
        self.files.borrow_mut().clear();
        self.subdirs.borrow_mut().clear();
    }
}

pub enum PathSegment {
    This,
    Parent,
    SubDirectory(String),
    Root,
}

impl PathSegment {
    pub fn parse_path_segment(path: String) -> Vec<PathSegment> {
        let mut segments: Vec<PathSegment> = Vec::new();
        let mut current: String = String::new();

        let mut first = true;

        for c in path.chars() {
            if c == '/' {
                if first {
                    segments.push(PathSegment::Root);
                }
                first = false;

                if current == "." {
                    segments.push(PathSegment::This);
                } else if current == ".." {
                    segments.push(PathSegment::Parent);
                } else {
                    segments.push(PathSegment::SubDirectory(current));
                }
                current = String::new();
            } else {
                current.push(c);
            }
        }

        if current == "." {
            segments.push(PathSegment::This);
        } else if current == ".." {
            segments.push(PathSegment::Parent);
        } else {
            segments.push(PathSegment::SubDirectory(current));
        }

        segments
    }
}
